// $Id: ETMove.cpp,v 1.1 2002/05/08 17:22:40 mrq Exp $

#include "ETMove.h"

#include "Application.h"
#include <sstream>

ETMove::ETMove (World *world) : EntityTool (world), m_Entity (NULL)
{}

ETMove::~ETMove()
{
   if (m_Entity) m_Entity->Unref();
}

/// Apply tool at (x,z)
void
ETMove::Apply3D(Ark::Vector3 newpos, const Ark::Collision &col)
{
   if (col.GetType() == Ark::Collision::ENTITY && m_Entity == NULL)
   {
      m_Entity = static_cast<Entity*>(col.m_Entity);
      m_Entity->Ref();
      m_NewPos = m_Entity->m_MState.m_Position;
      return;
   }
   else
   {
      std::cerr << "Apply3D " <<  col.m_Pos << "\n";
      m_NewPos = col.m_Pos;
      m_Entity->SetPosition(m_NewPos);
   }
}

/// Update or reset tool
void
ETMove::Update()
{
   if (m_Entity)
   {
      m_Entity->Unref();
      m_Entity = NULL;
   }
}

/// Create a command with a correct state.
Command*
ETMove::CreateCommand()
{
   if (m_Entity == NULL)
      return NULL;
   else
   {
      Command *c = new ECMove (g_Application->GetWorld(), m_Entity, m_NewPos);
      c->Execute();
      return c;
   }
}

///////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////

ECMove::ECMove(World *world, Entity *entity, Ark::Vector3 newpos) : 
   m_World (world),
   m_Entity (entity)
{
   assert (m_Entity != 0);
   m_Entity->Ref();

   m_NewPos = newpos;
   m_OldPos = m_Entity->m_MState.m_Position;
}

ECMove::~ECMove()
{
   if (m_Entity)
      m_Entity->Unref();
}

/// Return the command type.
std::string
ECMove::Type() const
{
   return "Move entity";
}

/// Execute this command.
void
ECMove::Execute()
{
   std::cerr << "Setting entity position to " << m_NewPos << "\n";
   m_Entity->SetPosition(m_NewPos); 
}

/// Undo the action done by this command.
void
ECMove::Unexecute()
{
   m_Entity->SetPosition(m_OldPos);
}
