// $Id: GLTerrainView.cpp,v 1.14 2002/10/04 20:59:10 teuf Exp $

#include <vector>
#include <string>
#include <iostream>
#include <algorithm>

#ifdef WIN32
# include <windows.h>
#endif
#include <gtk/gtksignal.h>

#include <Modules/Renderer/GLRenderer.h>

#include "../Application.h"
#include "../World.h"
#include "GLTerrainView.h"


#if 0 // not used
static void terrain_SetLight(Ark::Vector3 light0Pos)
{
   GLfloat light0Amb[4] = {0.20, 0.20, 0.20, 1.00};
   GLfloat light0Diff[4] = {1.00, 1.00, 1.00, 1.00};
   GLfloat light0Spec[4] = {1.00, 1.00, 1.00, 1.00};
   GLfloat light0SpotDir[3] = {0.00, 1.00, 1.00};
   
   GLfloat light0SpotExp = 0.00;
   GLfloat light0SpotCutoff = 180.00;
   GLfloat light0Atten0 = 1.00;
   GLfloat light0Atten1 = 0.00;
   GLfloat light0Atten2 = 0.00;
   
   glEnable(GL_LIGHT0);
   glLightfv(GL_LIGHT0, GL_POSITION, &light0Pos.X);
   glLightfv(GL_LIGHT0, GL_AMBIENT, light0Amb);
   glLightfv(GL_LIGHT0, GL_DIFFUSE, light0Diff);
   glLightfv(GL_LIGHT0, GL_SPECULAR, light0Spec);
   glLightfv(GL_LIGHT0, GL_SPOT_DIRECTION, light0SpotDir);
   glLightf(GL_LIGHT0, GL_SPOT_EXPONENT, light0SpotExp);
   glLightf(GL_LIGHT0, GL_SPOT_CUTOFF, light0SpotCutoff);
   glLightf(GL_LIGHT0, GL_CONSTANT_ATTENUATION, light0Atten0);
   glLightf(GL_LIGHT0, GL_LINEAR_ATTENUATION, light0Atten1);
   glLightf(GL_LIGHT0, GL_QUADRATIC_ATTENUATION, light0Atten2);
}
#endif

// =========================================================================


GLTerrainView::GLTerrainView(TerrainView* terrain) : 
  GLView(),
  m_Terrain (terrain)
{
   m_BBoxVB.SetFormat (Ark::VertexBuffer::VB_HAS_COORD);
   m_BBoxVB.Resize (18);
}

GLTerrainView::~GLTerrainView ()
{
}


// FIXME: put the following three functions somewhere in the engine, 'cause
// they are used both by the client and the world editor.
void RenderBBox (Ark::Renderer *renderer,
		 const Ark::BBox &bb,
		 const Ark::Material &mat,
		 Ark::VertexBuffer &vb)
{
   vb.Resize (10);
   
   Ark::Vector3 vs[8];
   bb.GetCorners (vs);
   int strips[][11] = { {1,3,0,2,6,4,7,5,1,3, -1},
			{7,1,6,0, -1},
			{3,5,2,4, -1} };
   
   for (int i = 0; i < 3; i++)
   {
      int j;
      for (j = 0; strips[i][j] != -1; j++)
	 vb.Coord(j) = vs[strips[i][j]];

      renderer->RenderBlock (mat, Ark::PRIM_TRIANGLE_STRIP, vb, j);
   }
}

static
Ark::Vector3 
UnProject (scalar sX, scalar sY, scalar mz,
	   Ark::Matrix44 invmat)
{
   Ark::Vector3 R = invmat.Transform(Ark::Vector3 (sX, sY, mz));
   scalar w = (sX * invmat.M(0,3) +
	       sY * invmat.M(1,3) +
	       mz * invmat.M(2,3) +
	       invmat.M(3,3));
   
   R.Scale (1.0/w);
   return R;
}
  

Ark::Vector3
GLTerrainView::TraceMousePos (const Ark::Vector3 &pov,
			      int mx, int my)
{      
   Ark::Matrix44 projmat, modelmat;
   int w = GTK_WIDGET(m_Area)->allocation.width;
   int h = GTK_WIDGET(m_Area)->allocation.height;
   
   glGetFloatv (GL_PROJECTION_MATRIX, &projmat.M(0,0));
   glGetFloatv (GL_MODELVIEW_MATRIX, &modelmat.M(0,0));
   
   modelmat.Multiply (projmat);
   modelmat.Invert();
   
   scalar sX = 2.0 * (scalar(mx) / scalar(w)) - 1.0;
   scalar sY = 2.0 * (scalar(h-my) / scalar(h)) - 1.0;
   
   Ark::Vector3 lookat = UnProject (sX, sY, 1.0, modelmat);
   
   Ark::Vector3 dir = lookat - pov;
   dir.Scale (10000.0);
   
   lookat = pov + dir;
   std::vector<Ark::Collision> collisions;
   m_Terrain->GetWorld()->RayTrace (Ark::Ray(pov, lookat),
				    Ark::Collision::POSITION |
				    Ark::Collision::ENTITY |
				    Ark::Collision::WORLD,
				    collisions);
   if (collisions.size())
   {
      m_LastHit = collisions[0];
      return collisions[0].m_Pos;
   }
   else
   {
      m_LastHit.m_Flags = 0;
      m_LastHit.m_Entity = NULL;
   }
   
   return Ark::Vector3 (0.0, 0.0, 0.0);
}


bool
GLTerrainView::Render()
{
  int w = GTK_WIDGET(m_Area)->allocation.width;
  int h = GTK_WIDGET(m_Area)->allocation.height;

  m_Renderer->SetViewport(0, 0, w, h);
  glClear(GL_DEPTH_BUFFER_BIT|GL_COLOR_BUFFER_BIT);

  // Call to cwterrain -> height field Render functions
  m_Terrain->Render();

  // Render bbox for selected entity.
  if(m_Terrain->GetWorld()->GetEntities().empty())
     return true;

  Ark::Material material("bbox/material");

  material.m_Flags = Ark::MATERIAL_HAS_PASS1;
  material.m_Passes[0].m_Flags |= Ark::PASS_HAS_BLENDING;
  material.m_Passes[0].m_BlendColor = Ark::Color (0.5f, 0.5f, 1.0f, 0.8f);
  material.m_Passes[0].m_BlendSrc = Ark::BLEND_ONE_MINUS_SRC_ALPHA;
  material.m_Passes[0].m_BlendDst = Ark::BLEND_SRC_ALPHA;

  RenderBBox (m_Renderer,
	      m_Terrain->GetWorld()->GetEntities()[0]->GetBBox(),
	      material, m_BBoxVB);

  return true;
}



static bool
ValidVector(World* w, const Ark::Vector3& vec)
{
  const int mx = w->m_SizeX;
  if ((vec.X < 0) || (mx <= vec.X)) return false;
  
  const int mz = w->m_SizeZ;
  if ((vec.Z < 0) || (mz <= vec.Z)) return false;

  return true;
}


bool
GLTerrainView::Motion(int x, int y, int state, int lastx, int lasty)
{
  bool redraw= true;
  
  const int Left = GDK_BUTTON1_MASK;
  const int Middle = GDK_BUTTON2_MASK;
  const int Right = GDK_BUTTON3_MASK;
  const int CtrlRight = (Right | GDK_CONTROL_MASK);
  const int ShiftRight = (Right | GDK_SHIFT_MASK);

  const scalar w = m_Width;
  const scalar h = m_Height;
  
  // Zooming using Control+Right
  if ((state & CtrlRight) == CtrlRight )
  {
    // Distance change
    m_Terrain->Zoom( 8.0 * (lasty - y) / h );

  }
  // Angle changes using Shift+Right or Middle button
  else if ( (state & Middle) || (state & ShiftRight) == ShiftRight )
  {
    // Vertical angle change
    m_Terrain->Rotate( 180.0 * (lastx - x) / w );
    
    // Pitch angle change
    m_Terrain->Pitch( 90.0 * (lasty - y) / h );
    
  }
  // Moving using right button
  else if (state & Right) 
   m_Terrain->Drag( (x-lastx) / w, (y-lasty) / h);
    
  // Use default action for this move
  else if (state & Left)
  {
    Ark::Vector3 v = GetGLPos(x, y);

    if (!ValidVector(m_Terrain->GetWorld(), v))
      redraw = false;
    else 
    {
       TraceMousePos(m_Terrain->GetCamera()->m_PointOfView, x, y);
       m_Terrain->Change(v, m_LastHit);
    }

  }
  else
    redraw = false;


  if (redraw)
  {
    Redraw();
    return true;
  }

  return false;
}


bool
GLTerrainView::ButtonPress(int x, int y, int button, int state)
{
   if (button == 1)
   {
      Ark::Vector3 v = GetGLPos(x, y);
      TraceMousePos(m_Terrain->GetCamera()->m_PointOfView, x, y);

      if (!ValidVector(m_Terrain->GetWorld(), v))
	 return false;
      
      m_Terrain->Change(v, m_LastHit);
      
      Redraw();
   }
   
   m_LX = x;
   m_LY = y;

   return true;
}


bool
GLTerrainView::ButtonRelease(int x, int y, int button, int state)
{

   // Distance on wheel !
   if ((button == 4) || (button == 5))
   {
     
     m_Terrain->Zoom( (button == 4) ? 1.0 : -1.0 );
     Redraw();
     
     return true;
   }
   
   if (button == 1)
   {
     m_Terrain->UpdateTool();
   }

   return false;
}


bool
GLTerrainView::Change(int fromx, int fromy, int tox, int toy)
{
  /*
  World* world = m_Terrain->GetWorld();
  
  Ark::Vector3 vfrom = GetGLPos(fromx, fromy);
  if (!ValidVector(world, vfrom)) return false;
        
  Ark::Vector3 vto = GetGLPos(tox, toy);
  if (!ValidVector(world, vto)) return false;

  Ark::Vector3 diff = vto - vfrom;

  const int length = static_cast<int>( diff.GetMagnitude() ) + 2;
  const scalar scale = 1.0 / static_cast<scalar>( length );

  diff.Scale( scale );

  for (int i = 0 ; i < length ; ++i) {
    m_Terrain->Change( vfrom );
    vfrom += diff;
  }
  */
  return true;
}
